/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Text Conferencing Test Application
 * 
 * Copyright (C) 2005, Joshua Colp
 *
 * Joshua Colp  <joshnet@nbnet.nb.ca>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "asterisk.h"

#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/app.h"
#include "asterisk/messaging.h"
#include "asterisk/astobj.h"

static char *tdesc = "Text Conferencing Test Application";
static char *app = "Chatroom";
static char *synopsis = 
"Text Conferencing Test Application.";
static char *descrip = "This application allows you to hold text based conferences using the new messaging architecture.\n";

/* Participant structure */
struct participant {
	struct ast_channel *owner;
	struct participant *next;
	struct ast_callerid cid;
};

/* Chatroom structure */
struct chatroom {
	ASTOBJ_COMPONENTS(struct chatroom);
	int locked;
	int participants_num;
	struct participant *participants;
};

static struct chatroom_list {
	ASTOBJ_CONTAINER_COMPONENTS(struct chatroom);
} chatrooms;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static void participant_destroy(struct participant *participant)
{
	participant->owner = NULL;
	free(participant);
	return;
}

static void chatroom_destroy(struct chatroom *chatroom)
{
	struct participant *participant = NULL, *participant_next = NULL;
	/* Go through the participants list and kill them all too */
	participant = chatroom->participants;
	while (participant) {
		if (participant->owner)
			ast_hangup(participant->owner);
		participant_next = participant->next;
		participant_destroy(participant);
		participant = participant_next;
	}
	free(chatroom);
	return;
}

static void chatroom_message(char *name, char *text)
{
        struct chatroom *chatroom = NULL;
        struct participant *participant = NULL;
        chatroom = ASTOBJ_CONTAINER_FIND(&chatrooms, name);
        if (strlen(text) == 0 || !chatroom) {
                return;
        }
        /* Now iterate through the list... */
        participant = chatroom->participants;
        while (participant) {
                ast_copy_string(participant->owner->cid.cid_num, name, sizeof(participant->owner->cid.cid_num));
                ast_sendtext(participant->owner, text);
                participant = participant->next;
        }
        return;
}

static int chatroom_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct localuser *u;
	struct chatroom *chatroom = NULL;
	struct participant *participant = NULL, *prev_part = NULL, *next_part = NULL;
	char text[1024] = "";
	int done = 0;
	char buf[1024] = "";

	/* Let's first make sure this is a text based channel paleez */
	if (chan->content != 1) {
		ast_log(LOG_NOTICE, "Trying to call a text-only application on a voice based channel is a big nono\n");
		return -1;
	}

	LOCAL_USER_ADD(u);

	/* See if the chatroom already exists */
	chatroom = ASTOBJ_CONTAINER_FIND(&chatrooms, chan->exten);
	if (!chatroom) {
		/* The chatroom does not exist yet - so create it. */
		chatroom = (struct chatroom *)malloc(sizeof(struct chatroom));
		if (!chatroom) {
			LOCAL_USER_REMOVE(u);
			return -1;
		}
		memset(chatroom, 0, sizeof(struct chatroom));
		ASTOBJ_INIT(chatroom);
		ast_copy_string(chatroom->name, chan->exten, sizeof(chatroom->name));
		chatroom->locked = 0; /* By default have it unlocked */
		chatroom->participants_num = 0; /* No participants yet */
		chatroom->participants = NULL; /* No participants! */
		ASTOBJ_CONTAINER_LINK(&chatrooms, chatroom);
		ASTOBJ_UNREF(chatroom, chatroom_destroy);
		chatroom = ASTOBJ_CONTAINER_FIND(&chatrooms, chan->exten);
	}

	/* Okay check to make sure we can enter the room! */
	if (chatroom->locked == 1) {
		ast_sendtext(chan, "This chatroom is locked.");
		LOCAL_USER_REMOVE(u);
		return -1;
	}

	/* Okay we can enter the chatroom... add us in */
	participant = (struct participant *)malloc(sizeof(struct participant));
	if (!participant) {
		ast_sendtext(chan, "Could not join chatroom - memory error.");
		LOCAL_USER_REMOVE(u);
		return -1;
	}
	memset(participant, 0, sizeof(struct participant));
	participant->owner = chan;
	participant->next = chatroom->participants;
	/* Copy over our callerid information... */

	chatroom->participants = participant;
	chatroom->participants_num += 1;

	/* Announce our join */
	snprintf(text, sizeof(text), "%s has joined the chatroom", chan->cid.cid_name);
	chatroom_message(chan->exten, text);

	while (!done) {
		struct ast_frame *f;
		if (ast_check_hangup(chan))
			break;
		res = ast_waitfor(chan, 3600);
		f = ast_read(chan);
		if (f != NULL) {
			if (f->frametype == AST_FRAME_CONTROL && f->subclass == AST_CONTROL_HANGUP)
				done = 1;
			else if (f->frametype == AST_FRAME_TEXT) {
				/* We have text... let's use the unb0rken method of getting it */
				memset(buf, 0, sizeof(buf));
				strncpy(buf, f->data, f->datalen);
				snprintf(text, sizeof(text), "%s says: %s", chan->cid.cid_name, buf);
				chatroom_message(chan->exten, text);
			}
			ast_frfree(f);
		}
	}

	/* Announce our departure */
	snprintf(text, sizeof(text), "%s has left the chatroom", chan->cid.cid_name);
	chatroom_message(chan->exten, text);

	/* Get rid of ourselves based on owner */
	chatroom = ASTOBJ_CONTAINER_FIND(&chatrooms, chan->exten);
	participant = chatroom->participants;
	next_part = NULL;
	prev_part = NULL;
	while (participant) {
		next_part = participant->next;
		if (participant->owner == chan) {
			/* Destroy us... */
			if (prev_part)
				prev_part->next = next_part;
			free(participant);
			participant = NULL;
		}
		prev_part = participant;
		participant = next_part;
	}

	/* If no more participants in the chatroom, destroy it */
	if (chatroom->participants_num == 0) {
		/* Destroy it... */
		chatroom = ASTOBJ_CONTAINER_FIND_UNLINK(&chatrooms, chan->exten);
		ASTOBJ_UNREF(chatroom, chatroom_destroy);		
	}

	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
        ASTOBJ_CONTAINER_DESTROYALL(&chatrooms, chatroom_destroy);
        ASTOBJ_CONTAINER_DESTROY(&chatrooms);
	return ast_unregister_application(app);
}

int load_module(void)
{
	ASTOBJ_CONTAINER_INIT(&chatrooms);
	return ast_register_application(app, chatroom_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
